@@ -25,6 +25,7 @@
 #include "builtins.h"
 #include "shell.h"
 
+int
 necho_builtin (list)
 WORD_LIST *list;
 {
@@ -37,7 +38,7 @@ WORD_LIST *list;
 char *necho_doc[] = {
 	"Display arguments.",
 	"",
-	"Print the arguments to the standard ouput separated",
+	"Print the arguments to the standard output separated",
 	"by space characters and terminated with a newline.",
 	(char *)NULL
 };
