@@ -1,7 +1,7 @@
 /* trap.c -- Not the trap command, but useful functions for manipulating
    those objects.  The trap command is in builtins/trap.def. */
 
-/* Copyright (C) 1987-2010 Free Software Foundation, Inc.
+/* Copyright (C) 1987-2020 Free Software Foundation, Inc.
 
    This file is part of GNU Bash, the Bourne Again SHell.
 
@@ -33,10 +33,14 @@
 
 #include "bashintl.h"
 
+#include <signal.h>
+
 #include "trap.h"
 
 #include "shell.h"
+#include "execute_cmd.h"
 #include "flags.h"
+#include "parser.h"
 #include "input.h"	/* for save_token_state, restore_token_state */
 #include "jobs.h"
 #include "signames.h"
@@ -44,6 +48,11 @@
 #include "builtins/common.h"
 #include "builtins/builtext.h"
 
+#if defined (READLINE)
+#  include <readline/readline.h>
+#  include "bashline.h"
+#endif
+
 #ifndef errno
 extern int errno;
 #endif
@@ -65,34 +74,31 @@ extern int errno;
    assumes this. */
 static int sigmodes[BASH_NSIG];
 
-static void free_trap_command __P((int));
-static void change_signal __P((int, char *));
+static void free_trap_command (int);
+static void change_signal (int, char *);
 
-static void get_original_signal __P((int));
+static int _run_trap_internal (int, char *);
 
-static int _run_trap_internal __P((int, char *));
+static void free_trap_string (int);
+static void reset_signal (int);
+static void restore_signal (int);
+static void reset_or_restore_signal_handlers (sh_resetsig_func_t *);
 
-static void free_trap_string __P((int));
-static void reset_signal __P((int));
-static void restore_signal __P((int));
-static void reset_or_restore_signal_handlers __P((sh_resetsig_func_t *));
+static void trap_if_untrapped (int, char *);
 
 /* Variables used here but defined in other files. */
-extern int last_command_exit_value;
-extern int line_number;
+extern procenv_t alrmbuf;
+
+extern volatile int from_return_trap;
+extern int waiting_for_child;
 
-extern char *this_command_name;
-extern sh_builtin_func_t *this_shell_builtin;
-extern procenv_t wait_intr_buf;
-extern int return_catch_flag, return_catch_value;
-extern int subshell_level;
 extern WORD_LIST *subst_assign_varlist;
 
 /* The list of things to do originally, before we started trapping. */
 SigHandler *original_signals[NSIG];
 
 /* For each signal, a slot for a string, which is a command to be
-   executed when that signal is recieved.  The slot can also contain
+   executed when that signal is received.  The slot can also contain
    DEFAULT_SIG, which means do whatever you were going to do before
    you were so rudely interrupted, or IGNORE_SIG, which says ignore
    this signal. */
@@ -113,6 +119,11 @@ int trap_saved_exit_value;
 /* The (trapped) signal received while executing in the `wait' builtin */
 int wait_signal_received;
 
+int trapped_signal_received;
+
+/* Set to 1 to suppress the effect of `set v' in the DEBUG trap. */
+int suppress_debug_trap_verbose = 0;
+
 #define GETORIGSIG(sig) \
   do { \
     original_signals[sig] = (SigHandler *)set_signal_handler (sig, SIG_DFL); \
@@ -174,9 +185,11 @@ initialize_traps ()
       GETORIGSIG (SIGTERM);
       sigmodes[SIGTERM] |= SIG_SPECIAL;
     }
+
+  get_original_tty_job_signals ();
 }
 
-#ifdef INCLUDE_UNUSED
+#ifdef DEBUG
 /* Return a printable representation of the trap handler for SIG. */
 static char *
 trap_handler_string (sig)
@@ -225,6 +238,23 @@ decode_signal (string, flags)
   if (legal_number (string, &sig))
     return ((sig >= 0 && sig < NSIG) ? (int)sig : NO_SIG);
 
+#if defined (SIGRTMIN) && defined (SIGRTMAX)
+  if (STREQN (string, "SIGRTMIN+", 9) || ((flags & DSIG_NOCASE) && strncasecmp (string, "SIGRTMIN+", 9) == 0))
+    {
+      if (legal_number (string+9, &sig) && sig >= 0 && sig <= SIGRTMAX - SIGRTMIN)
+	return (SIGRTMIN + sig);
+      else
+	return NO_SIG;
+    }
+  else if (STREQN (string, "RTMIN+", 6) || ((flags & DSIG_NOCASE) && strncasecmp (string, "RTMIN+", 6) == 0))
+    {
+      if (legal_number (string+6, &sig) && sig >= 0 && sig <= SIGRTMAX - SIGRTMIN)
+	return (SIGRTMIN + sig);
+      else
+	return NO_SIG;
+    }
+#endif /* SIGRTMIN && SIGRTMAX */
+
   /* A leading `SIG' may be omitted. */
   for (sig = 0; sig < BASH_NSIG; sig++)
     {
@@ -232,7 +262,7 @@ decode_signal (string, flags)
       if (name == 0 || name[0] == '\0')
 	continue;
 
-      /* Check name without the SIG prefix first case sensitivly or
+      /* Check name without the SIG prefix first case sensitively or
 	 insensitively depending on whether flags includes DSIG_NOCASE */
       if (STREQN (name, "SIG", 3))
 	{
@@ -267,8 +297,11 @@ void
 run_pending_traps ()
 {
   register int sig;
-  int old_exit_value, *token_state;
+  int old_exit_value, x;
+  int old_running;
   WORD_LIST *save_subst_varlist;
+  HASH_TABLE *save_tempenv;
+  sh_parser_state_t pstate;
 #if defined (ARRAY_VARS)
   ARRAY *ps;
 #endif
@@ -276,13 +309,33 @@ run_pending_traps ()
   if (catch_flag == 0)		/* simple optimization */
     return;
 
-  catch_flag = 0;
+  if (running_trap > 0)
+    {
+#if defined (DEBUG)
+      internal_warning ("run_pending_traps: recursive invocation while running trap for signal %d", running_trap-1);
+#endif
+#if defined (SIGWINCH)
+      if (running_trap == SIGWINCH+1 && pending_traps[SIGWINCH])
+	return;			/* no recursive SIGWINCH trap invocations */
+#endif
+      /* could check for running the trap handler for the same signal here
+	 (running_trap == sig+1) */
+      if (evalnest_max > 0 && evalnest > evalnest_max)
+	{
+	  internal_error (_("trap handler: maximum trap handler level exceeded (%d)"), evalnest_max);
+	  evalnest = 0;
+	  jump_to_top_level (DISCARD);
+	}
+    }
+
+  catch_flag = trapped_signal_received = 0;
 
   /* Preserve $? when running trap. */
-  old_exit_value = last_command_exit_value;
+  trap_saved_exit_value = old_exit_value = last_command_exit_value;
 #if defined (ARRAY_VARS)
   ps = save_pipestatus_array ();
 #endif
+  old_running = running_trap;
 
   for (sig = 1; sig < NSIG; sig++)
     {
@@ -290,31 +343,55 @@ run_pending_traps ()
 	 while (pending_traps[sig]--) instead of the if statement. */
       if (pending_traps[sig])
 	{
-#if defined (HAVE_POSIX_SIGNALS)
-	  sigset_t set, oset;
+	  if (running_trap == sig+1)
+	    /*continue*/;
 
-	  sigemptyset (&set);
-	  sigemptyset (&oset);
-
-	  sigaddset (&set, sig);
-	  sigprocmask (SIG_BLOCK, &set, &oset);
-#else
-#  if defined (HAVE_BSD_SIGNALS)
-	  int oldmask = sigblock (sigmask (sig));
-#  endif
-#endif /* HAVE_POSIX_SIGNALS */
+	  running_trap = sig + 1;
 
 	  if (sig == SIGINT)
 	    {
-	      run_interrupt_trap ();
-	      CLRINTERRUPT;
+	      pending_traps[sig] = 0;	/* XXX */
+	      /* We don't modify evalnest here, since run_interrupt_trap() calls
+		 _run_trap_internal, which does. */
+	      run_interrupt_trap (0);
+	      CLRINTERRUPT;	/* interrupts don't stack */
 	    }
 #if defined (JOB_CONTROL) && defined (SIGCHLD)
 	  else if (sig == SIGCHLD &&
 		   trap_list[SIGCHLD] != (char *)IMPOSSIBLE_TRAP_HANDLER &&
 		   (sigmodes[SIGCHLD] & SIG_INPROGRESS) == 0)
 	    {
-	      run_sigchld_trap (pending_traps[sig]);	/* use as counter */
+	      sigmodes[SIGCHLD] |= SIG_INPROGRESS;
+	      /* We modify evalnest here even though run_sigchld_trap can run
+		 the trap action more than once */
+	      evalnest++;
+	      x = pending_traps[sig];
+	      pending_traps[sig] = 0;
+	      run_sigchld_trap (x);	/* use as counter */
+	      running_trap = 0;
+	      evalnest--;
+	      sigmodes[SIGCHLD] &= ~SIG_INPROGRESS;
+	      /* continue here rather than reset pending_traps[SIGCHLD] below in
+		 case there are recursive calls to run_pending_traps and children
+		 have been reaped while run_sigchld_trap was running. */
+	      continue;
+	    }
+	  else if (sig == SIGCHLD &&
+		   trap_list[SIGCHLD] == (char *)IMPOSSIBLE_TRAP_HANDLER &&
+		   (sigmodes[SIGCHLD] & SIG_INPROGRESS) != 0)
+	    {
+	      /* This can happen when run_pending_traps is called while
+		 running a SIGCHLD trap handler. */
+	      running_trap = 0;
+	      /* want to leave pending_traps[SIGCHLD] alone here */
+	      continue;					/* XXX */
+	    }
+	  else if (sig == SIGCHLD && (sigmodes[SIGCHLD] & SIG_INPROGRESS))
+	    {
+	      /* whoops -- print warning? */
+	      running_trap = 0;		/* XXX */
+	      /* want to leave pending_traps[SIGCHLD] alone here */
+	      continue;
 	    }
 #endif
 	  else if (trap_list[sig] == (char *)DEFAULT_SIG ||
@@ -344,26 +421,32 @@ run_pending_traps ()
 	    }
 	  else
 	    {
-	      token_state = save_token_state ();
+	      /* XXX - should we use save_parser_state/restore_parser_state? */
+	      save_parser_state (&pstate);
 	      save_subst_varlist = subst_assign_varlist;
 	      subst_assign_varlist = 0;
+	      save_tempenv = temporary_env;
+	      temporary_env = 0;	/* traps should not run with temporary env */
 
-	      parse_and_execute (savestring (trap_list[sig]), "trap", SEVAL_NONINT|SEVAL_NOHIST|SEVAL_RESETLINE);
-	      restore_token_state (token_state);
-	      free (token_state);
+#if defined (JOB_CONTROL)
+	      save_pipeline (1);	/* XXX only provides one save level */
+#endif
+	      /* XXX - set pending_traps[sig] = 0 here? */
+	      pending_traps[sig] = 0;
+	      evalnest++;
+	      evalstring (savestring (trap_list[sig]), "trap", SEVAL_NONINT|SEVAL_NOHIST|SEVAL_RESETLINE);
+	      evalnest--;
+#if defined (JOB_CONTROL)
+	      restore_pipeline (1);
+#endif
 
 	      subst_assign_varlist = save_subst_varlist;
+	      restore_parser_state (&pstate);
+	      temporary_env = save_tempenv;
 	    }
 
-	  pending_traps[sig] = 0;
-
-#if defined (HAVE_POSIX_SIGNALS)
-	  sigprocmask (SIG_SETMASK, &oset, (sigset_t *)NULL);
-#else
-#  if defined (HAVE_BSD_SIGNALS)
-	  sigsetmask (oldmask);
-#  endif
-#endif /* POSIX_VERSION */
+	  pending_traps[sig] = 0;	/* XXX - move before evalstring? */
+	  running_trap = old_running;
 	}
     }
 
@@ -373,6 +456,17 @@ run_pending_traps ()
   last_command_exit_value = old_exit_value;
 }
 
+/* Set the private state variables noting that we received a signal SIG
+   for which we have a trap set. */
+void
+set_trap_state (sig)
+     int sig;
+{
+  catch_flag = 1;
+  pending_traps[sig]++;
+  trapped_signal_received = sig;
+}
+    
 sighandler
 trap_handler (sig)
      int sig;
@@ -401,24 +495,85 @@ trap_handler (sig)
       set_signal_handler (sig, trap_handler);
 #endif /* MUST_REINSTALL_SIGHANDLERS */
 
-      catch_flag = 1;
-      pending_traps[sig]++;
+      set_trap_state (sig);
 
-      if (interrupt_immediately && this_shell_builtin && (this_shell_builtin == wait_builtin))
+      if (this_shell_builtin && (this_shell_builtin == wait_builtin))
 	{
 	  wait_signal_received = sig;
-	  longjmp (wait_intr_buf, 1);
+	  if (waiting_for_child && wait_intr_flag)
+	    sh_longjmp (wait_intr_buf, 1);
 	}
 
-      if (interrupt_immediately)
-	run_pending_traps ();
+#if defined (READLINE)
+      /* Set the event hook so readline will call it after the signal handlers
+	 finish executing, so if this interrupted character input we can get
+	 quick response. */
+      if (RL_ISSTATE (RL_STATE_SIGHANDLER))
+        bashline_set_event_hook ();
+#endif
 
       errno = oerrno;
     }
-
+  
   SIGRETURN (0);
 }
 
+int
+next_pending_trap (start)
+     int start;
+{
+  register int i;
+
+  for (i = start; i < NSIG; i++)
+    if (pending_traps[i])
+      return i;
+  return -1;
+}
+
+int
+first_pending_trap ()
+{
+  return (next_pending_trap (1));
+}
+
+/* Return > 0 if any of the "real" signals (not fake signals like EXIT) are
+   trapped. */
+int
+any_signals_trapped ()
+{
+  register int i;
+
+  for (i = 1; i < NSIG; i++)
+    if (sigmodes[i] & SIG_TRAPPED)
+      return i;
+  return -1;
+}
+
+void
+clear_pending_traps ()
+{
+  register int i;
+
+  for (i = 1; i < NSIG; i++)
+    pending_traps[i] = 0;
+}
+
+void
+check_signals ()
+{
+  CHECK_ALRM;		/* set by the read builtin */
+  QUIT;
+}
+
+/* Convenience functions the rest of the shell can use */
+void
+check_signals_and_traps ()
+{
+  check_signals ();
+
+  run_pending_traps ();
+}
+
 #if defined (JOB_CONTROL) && defined (SIGCHLD)
 
 #ifdef INCLUDE_UNUSED
@@ -454,8 +609,33 @@ set_impossible_sigchld_trap ()
   change_signal (SIGCHLD, (char *)IMPOSSIBLE_TRAP_HANDLER);
   sigmodes[SIGCHLD] &= ~SIG_TRAPPED;	/* maybe_set_sigchld_trap checks this */
 }
+
+/* Act as if we received SIGCHLD NCHILD times and increment
+   pending_traps[SIGCHLD] by that amount.  This allows us to still run the
+   SIGCHLD trap once for each exited child. */
+void
+queue_sigchld_trap (nchild)
+     int nchild;
+{
+  if (nchild > 0)
+    {
+      catch_flag = 1;
+      pending_traps[SIGCHLD] += nchild;
+      trapped_signal_received = SIGCHLD;
+    }
+}
 #endif /* JOB_CONTROL && SIGCHLD */
 
+/* Set a trap for SIG only if SIG is not already trapped. */
+static inline void
+trap_if_untrapped (sig, command)
+     int sig;
+     char *command;
+{
+  if ((sigmodes[sig] & SIG_TRAPPED) == 0)
+    set_signal (sig, command);
+}
+
 void
 set_debug_trap (command)
      char *command;
@@ -463,6 +643,19 @@ set_debug_trap (command)
   set_signal (DEBUG_TRAP, command);
 }
 
+/* Separate function to call when functions and sourced files want to restore
+   the original version of the DEBUG trap before returning.  Unless the -T
+   option is set, source and shell function execution save the old debug trap
+   and unset the trap.  If the function or sourced file changes the DEBUG trap,
+   SIG_TRAPPED will be set and we don't bother restoring the original trap string.
+   This is used by both functions and the source builtin. */
+void
+maybe_set_debug_trap (command)
+     char *command;
+{
+  trap_if_untrapped (DEBUG_TRAP, command);
+}
+
 void
 set_error_trap (command)
      char *command;
@@ -470,6 +663,13 @@ set_error_trap (command)
   set_signal (ERROR_TRAP, command);
 }
 
+void
+maybe_set_error_trap (command)
+     char *command;
+{
+  trap_if_untrapped (ERROR_TRAP, command);
+}
+
 void
 set_return_trap (command)
      char *command;
@@ -477,6 +677,13 @@ set_return_trap (command)
   set_signal (RETURN_TRAP, command);
 }
 
+void
+maybe_set_return_trap (command)
+     char *command;
+{
+  trap_if_untrapped (RETURN_TRAP, command);
+}
+
 #ifdef INCLUDE_UNUSED
 void
 set_sigint_trap (command)
@@ -529,6 +736,8 @@ set_signal (sig, string)
      int sig;
      char *string;
 {
+  sigset_t set, oset;
+
   if (SPECIAL_TRAP (sig))
     {
       change_signal (sig, savestring (string));
@@ -559,9 +768,10 @@ set_signal (sig, string)
      environment in which it is safe to do so. */
   if ((sigmodes[sig] & SIG_NO_TRAP) == 0)
     {
-      set_signal_handler (sig, SIG_IGN);
+      BLOCK_SIGNAL (sig, set, oset);
       change_signal (sig, savestring (string));
       set_signal_handler (sig, trap_handler);
+      UNBLOCK_SIGNAL (oset);
     }
   else
     change_signal (sig, savestring (string));
@@ -598,7 +808,7 @@ change_signal (sig, value)
     sigmodes[sig] |= SIG_CHANGED;
 }
 
-static void
+void
 get_original_signal (sig)
      int sig;
 {
@@ -653,7 +863,11 @@ restore_default_signal (sig)
     return;
 
   /* If we aren't trapping this signal, don't bother doing anything else. */
-  if ((sigmodes[sig] & SIG_TRAPPED) == 0)
+  /* We special-case SIGCHLD and IMPOSSIBLE_TRAP_HANDLER (see above) as a
+     sentinel to determine whether or not disposition is reset to the default
+     while the trap handler is executing. */
+  if (((sigmodes[sig] & SIG_TRAPPED) == 0) &&
+      (sig != SIGCHLD || (sigmodes[sig] & SIG_INPROGRESS) == 0 || trap_list[sig] != (char *)IMPOSSIBLE_TRAP_HANDLER))
     return;
 
   /* Only change the signal handler for SIG if it allows it. */
@@ -728,11 +942,11 @@ run_exit_trap ()
       retval = trap_saved_exit_value;
       running_trap = 1;
 
-      code = setjmp (top_level);
+      code = setjmp_nosigs (top_level);
 
       /* If we're in a function, make sure return longjmps come here, too. */
       if (return_catch_flag)
-	function_code = setjmp (return_catch);
+	function_code = setjmp_nosigs (return_catch);
 
       if (code == 0 && function_code == 0)
 	{
@@ -749,6 +963,10 @@ run_exit_trap ()
 	retval = trap_saved_exit_value;
 
       running_trap = 0;
+#if defined (ARRAY_VARS)
+      array_dispose (ps);
+#endif
+
       return retval;
     }
 
@@ -762,9 +980,12 @@ void
 run_trap_cleanup (sig)
      int sig;
 {
+  /* XXX - should we clean up trap_list[sig] == IMPOSSIBLE_TRAP_HANDLER? */
   sigmodes[sig] &= ~(SIG_INPROGRESS|SIG_CHANGED);
 }
 
+#define RECURSIVE_SIG(s) (SPECIAL_TRAP(s) == 0)
+
 /* Run a trap command for SIG.  SIG is one of the signals the shell treats
    specially.  Returns the exit status of the executed trap command list. */
 static int
@@ -773,63 +994,102 @@ _run_trap_internal (sig, tag)
      char *tag;
 {
   char *trap_command, *old_trap;
-  int trap_exit_value, *token_state;
-  int save_return_catch_flag, function_code, flags;
+  int trap_exit_value;
+  volatile int save_return_catch_flag, function_code;
+  int old_modes, old_running, old_int;
+  int flags;
   procenv_t save_return_catch;
   WORD_LIST *save_subst_varlist;
+  HASH_TABLE *save_tempenv;
+  sh_parser_state_t pstate;
 #if defined (ARRAY_VARS)
   ARRAY *ps;
 #endif
 
+  old_modes = old_running = -1;
+
   trap_exit_value = function_code = 0;
+  trap_saved_exit_value = last_command_exit_value;
   /* Run the trap only if SIG is trapped and not ignored, and we are not
      currently executing in the trap handler. */
   if ((sigmodes[sig] & SIG_TRAPPED) && ((sigmodes[sig] & SIG_IGNORED) == 0) &&
       (trap_list[sig] != (char *)IMPOSSIBLE_TRAP_HANDLER) &&
+#if 1
+      /* Uncomment this to allow some special signals to recursively execute
+	 trap handlers. */
+      (RECURSIVE_SIG (sig) || (sigmodes[sig] & SIG_INPROGRESS) == 0))
+#else
       ((sigmodes[sig] & SIG_INPROGRESS) == 0))
+#endif
     {
       old_trap = trap_list[sig];
+      old_modes = sigmodes[sig];
+      old_running = running_trap;
+
       sigmodes[sig] |= SIG_INPROGRESS;
       sigmodes[sig] &= ~SIG_CHANGED;		/* just to be sure */
       trap_command =  savestring (old_trap);
 
       running_trap = sig + 1;
-      trap_saved_exit_value = last_command_exit_value;
+
+      old_int = interrupt_state;	/* temporarily suppress pending interrupts */
+      CLRINTERRUPT;
+
 #if defined (ARRAY_VARS)
       ps = save_pipestatus_array ();
 #endif
 
-      token_state = save_token_state ();
+      save_parser_state (&pstate);
       save_subst_varlist = subst_assign_varlist;
       subst_assign_varlist = 0;
+      save_tempenv = temporary_env;
+      temporary_env = 0;	/* traps should not run with temporary env */
+
+#if defined (JOB_CONTROL)
+      if (sig != DEBUG_TRAP)	/* run_debug_trap does this */
+	save_pipeline (1);	/* XXX only provides one save level */
+#endif
 
       /* If we're in a function, make sure return longjmps come here, too. */
       save_return_catch_flag = return_catch_flag;
       if (return_catch_flag)
 	{
 	  COPY_PROCENV (return_catch, save_return_catch);
-	  function_code = setjmp (return_catch);
+	  function_code = setjmp_nosigs (return_catch);
 	}
 
       flags = SEVAL_NONINT|SEVAL_NOHIST;
       if (sig != DEBUG_TRAP && sig != RETURN_TRAP && sig != ERROR_TRAP)
 	flags |= SEVAL_RESETLINE;
+      evalnest++;
       if (function_code == 0)
-	parse_and_execute (trap_command, tag, flags);
+        {
+	  parse_and_execute (trap_command, tag, flags);
+	  trap_exit_value = last_command_exit_value;
+        }
+      else
+        trap_exit_value = return_catch_value;
+      evalnest--;
 
-      restore_token_state (token_state);
-      free (token_state);
+#if defined (JOB_CONTROL)
+      if (sig != DEBUG_TRAP)	/* run_debug_trap does this */
+	restore_pipeline (1);
+#endif
 
       subst_assign_varlist = save_subst_varlist;
+      restore_parser_state (&pstate);
 
-      trap_exit_value = last_command_exit_value;
-      last_command_exit_value = trap_saved_exit_value;
 #if defined (ARRAY_VARS)
       restore_pipestatus_array (ps);
 #endif
-      running_trap = 0;
 
-      sigmodes[sig] &= ~SIG_INPROGRESS;
+      temporary_env = save_tempenv;
+
+      if ((old_modes & SIG_INPROGRESS) == 0)
+	sigmodes[sig] &= ~SIG_INPROGRESS;
+
+      running_trap = old_running;
+      interrupt_state = old_int;
 
       if (sigmodes[sig] & SIG_CHANGED)
 	{
@@ -841,6 +1101,8 @@ _run_trap_internal (sig, tag)
 #endif
 	    free (old_trap);
 	  sigmodes[sig] &= ~SIG_CHANGED;
+
+	  CHECK_TERMSIG;	/* some pathological conditions lead here */
 	}
 
       if (save_return_catch_flag)
@@ -849,7 +1111,12 @@ _run_trap_internal (sig, tag)
 	  return_catch_value = trap_exit_value;
 	  COPY_PROCENV (save_return_catch, return_catch);
 	  if (function_code)
-	    longjmp (return_catch, 1);
+	    {
+#if 0
+	      from_return_trap = sig == RETURN_TRAP;
+#endif
+	      sh_longjmp (return_catch, 1);
+	    }
 	}
     }
 
@@ -859,9 +1126,11 @@ _run_trap_internal (sig, tag)
 int
 run_debug_trap ()
 {
-  int trap_exit_value;
+  int trap_exit_value, old_verbose;
   pid_t save_pgrp;
+#if defined (PGRP_PIPE)
   int save_pipe[2];
+#endif
 
   /* XXX - question:  should the DEBUG trap inherit the RETURN trap? */
   trap_exit_value = 0;
@@ -877,8 +1146,13 @@ run_debug_trap ()
       stop_making_children ();
 #endif
 
+      old_verbose = echo_input_at_read;
+      echo_input_at_read = suppress_debug_trap_verbose ? 0 : echo_input_at_read;
+
       trap_exit_value = _run_trap_internal (DEBUG_TRAP, "debug trap");
 
+      echo_input_at_read = old_verbose;
+
 #if defined (JOB_CONTROL)
       pipeline_pgrp = save_pgrp;
       restore_pipeline (1);
@@ -886,8 +1160,9 @@ run_debug_trap ()
       close_pgrp_pipe ();
       restore_pgrp_pipe (save_pipe);
 #  endif
-      if (pipeline_pgrp > 0)
+      if (pipeline_pgrp > 0 && ((subshell_environment & (SUBSHELL_ASYNC|SUBSHELL_PIPE)) == 0))
 	give_terminal_to (pipeline_pgrp, 1);
+
       notify_and_cleanup ();
 #endif
       
@@ -897,7 +1172,7 @@ run_debug_trap ()
       if (debugging_mode && trap_exit_value == 2 && return_catch_flag)
 	{
 	  return_catch_value = trap_exit_value;
-	  longjmp (return_catch, 1);
+	  sh_longjmp (return_catch, 1);
 	}
 #endif
     }
@@ -932,23 +1207,42 @@ run_return_trap ()
 /* Run a trap set on SIGINT.  This is called from throw_to_top_level (), and
    declared here to localize the trap functions. */
 void
-run_interrupt_trap ()
+run_interrupt_trap (will_throw)
+     int will_throw;	/* from throw_to_top_level? */
 {
+  if (will_throw && running_trap > 0)
+    run_trap_cleanup (running_trap - 1);
+  pending_traps[SIGINT] = 0;	/* run_pending_traps does this */
+  catch_flag = 0;
   _run_trap_internal (SIGINT, "interrupt trap");
 }
 
 /* Free all the allocated strings in the list of traps and reset the trap
    values to the default.  Intended to be called from subshells that want
    to complete work done by reset_signal_handlers upon execution of a
-   subsequent `trap' command that changes a signal's disposition. */
+   subsequent `trap' command that changes a signal's disposition.  We need
+   to make sure that we duplicate the behavior of
+   reset_or_restore_signal_handlers and not change the disposition of signals
+   that are set to be ignored. */
 void
 free_trap_strings ()
 {
   register int i;
 
-  for (i = 0; i < BASH_NSIG; i++)
-    free_trap_string (i);
-  trap_list[DEBUG_TRAP] = trap_list[EXIT_TRAP] = trap_list[ERROR_TRAP] = trap_list[RETURN_TRAP] = (char *)NULL;
+  for (i = 0; i < NSIG; i++)
+    {
+      if (trap_list[i] != (char *)IGNORE_SIG)
+	free_trap_string (i);
+    }
+  for (i = NSIG; i < BASH_NSIG; i++)
+    {
+      /* Don't free the trap string if the subshell inherited the trap */
+      if ((sigmodes[i] & SIG_TRAPPED) == 0)
+	{
+	  free_trap_string (i);
+	  trap_list[i] = (char *)NULL;
+	}
+    }
 }
 
 /* Free a trap command string associated with SIG without changing signal
@@ -958,7 +1252,7 @@ free_trap_string (sig)
      int sig;
 {
   change_signal (sig, (char *)DEFAULT_SIG);
-  sigmodes[sig] &= ~SIG_TRAPPED;
+  sigmodes[sig] &= ~SIG_TRAPPED;		/* XXX - SIG_INPROGRESS? */
 }
 
 /* Reset the handler for SIG to the original value but leave the trap string
@@ -968,7 +1262,7 @@ reset_signal (sig)
      int sig;
 {
   set_signal_handler (sig, original_signals[sig]);
-  sigmodes[sig] &= ~SIG_TRAPPED;
+  sigmodes[sig] &= ~SIG_TRAPPED;		/* XXX - SIG_INPROGRESS? */
 }
 
 /* Set the handler signal SIG to the original and free any trap
@@ -991,7 +1285,7 @@ reset_or_restore_signal_handlers (reset)
   /* Take care of the exit trap first */
   if (sigmodes[EXIT_TRAP] & SIG_TRAPPED)
     {
-      sigmodes[EXIT_TRAP] &= ~SIG_TRAPPED;
+      sigmodes[EXIT_TRAP] &= ~SIG_TRAPPED;	/* XXX - SIG_INPROGRESS? */
       if (reset != reset_signal)
 	{
 	  free_trap_command (EXIT_TRAP);
@@ -1010,6 +1304,7 @@ reset_or_restore_signal_handlers (reset)
 	}
       else if (sigmodes[i] & SIG_SPECIAL)
 	(*reset) (i);
+      pending_traps[i] = 0;	/* XXX */
     }
 
   /* Command substitution and other child processes don't inherit the
@@ -1045,7 +1340,7 @@ restore_original_signals ()
 }
 
 /* If a trap handler exists for signal SIG, then call it; otherwise just
-   return failure. */
+   return failure.  Returns 1 if it called the trap handler. */
 int
 maybe_call_trap_handler (sig)
      int sig;
@@ -1056,7 +1351,7 @@ maybe_call_trap_handler (sig)
       switch (sig)
 	{
 	case SIGINT:
-	  run_interrupt_trap ();
+	  run_interrupt_trap (0);
 	  break;
 	case EXIT_TRAP:
 	  run_exit_trap ();
@@ -1084,6 +1379,13 @@ signal_is_trapped (sig)
   return (sigmodes[sig] & SIG_TRAPPED);
 }
 
+int
+signal_is_pending (sig)
+     int sig;
+{
+  return (pending_traps[sig]);
+}
+
 int
 signal_is_special (sig)
      int sig;
@@ -1106,16 +1408,46 @@ signal_is_hard_ignored (sig)
 }
 
 void
-set_signal_ignored (sig)
+set_signal_hard_ignored (sig)
      int sig;
 {
   sigmodes[sig] |= SIG_HARD_IGNORE;
   original_signals[sig] = SIG_IGN;
 }
 
+void
+set_signal_ignored (sig)
+     int sig;
+{
+  original_signals[sig] = SIG_IGN;
+}
+
 int
 signal_in_progress (sig)
      int sig;
 {
   return (sigmodes[sig] & SIG_INPROGRESS);
 }
+
+#if 0 /* TAG: bash-5.2 */
+int
+block_trapped_signals (maskp, omaskp)
+     sigset_t *maskp;
+     sigset_t *omaskp;
+{
+  int i;
+
+  sigemptyset (maskp);
+  for (i = 1; i < NSIG; i++)
+    if (sigmodes[i] & SIG_TRAPPED)
+      sigaddset (maskp, i);
+  return (sigprocmask (SIG_BLOCK, maskp, omaskp));
+}
+
+int
+unblock_trapped_signals (maskp)
+     sigset_t *maskp;
+{
+  return (sigprocmask (SIG_SETMASK, maskp, 0));
+}
+#endif
